// ffe.cc        -*- mode: c++ -*-
// Copyright  1999 Laboratoire de Biologie Informatique et Thorique.
// Author           : Patrick Gendron <gendrop@vertigo.IRO.UMontreal.CA>
// Created on       : Tue Apr 11 10:04:05 2000
// Modified by      : 
// Modified on      : 
// Update count     : 0
// Status           : Unknown.

#include <iostream.h>
#include <stdlib.h>
#include <vector.h>
#include <pair.h>
#include <algo.h>
#include <pthread.h>
#include <unistd.h>
#include <fstream.h>
#include <time.h>

#include "ffe.h"

#ifdef HAVE_LIBQT
#include <qapplication.h>
#include "DataWrap.h"
#include "ScatterPlotW.h"
#endif

extern "C"{
#include "version.h"
#include "proto.h"
#include "protovx.h"
#include "protowbx.h"
#include "protowx.h"
#include "squid.h" 
}

int __current_option;
const char* __shortopts = "dt:gkl:s:w:hH?vV";


// Version function -----------------------------------------------------------

void __version (void)
{
  cout << actual_name << " " << VERSION
       << " (" << __DATE__ << ")" << endl;
  cout << "using: pknot " << RELEASE << " " << RELEASEDATE << endl;
  cout << "       squid " << squid_version << " " << squid_date << endl;
}

// Usage function -------------------------------------------------------------

void __usage (void)
{
  cout << "usage: " << actual_name 
       << " [-dt:gklswhvV] [FASTA seq file] " << endl;
}

// Help function --------------------------------------------------------------

void __help (void)
{
  cout << "  -d        dinucleotides  randomize dinucleotides" << endl
       << "  -t<x>     codons         randomize codons with a shift of x base" << endl
       << "  -k        pseudoknot     allow pseudoknots" << endl
       << "  -g        ct             out of a ct file" << endl
       << "  -l [int]  overlap        overlap region size (default=150)" << endl
       << "  -s [int]  shuffle        amount of sequence shuffle in AFFE (default=500)" << endl
       << "  -w [int]  window         size of the window (default=200)" << endl
       << "  -h        help           print this help" << endl
       << "  -v        verbose        be verbose" << endl
       << "  -V        version        print the software version info" << endl;
}


// Thread functions -----------------------------------------------------------

#ifdef HAVE_LIBQT

QApplication *gQApp = NULL;
ScatterPlotW *scatter;
pthread_t theThread;

void UpdatePlot (void)
{
}

void *CreateQApplicationAux (void *);

QApplication *CreateQApplication (void)
{
  QApplication *tmpQApplication = gQApp;
  pthread_create (&theThread, NULL, CreateQApplicationAux, NULL);
  while (tmpQApplication == gQApp) sleep (1);
  return gQApp;
}

void *CreateQApplicationAux (void *)
{
  int bidon;
  char title[80];  
  gQApp = new QApplication(bidon, NULL);
  scatter = new ScatterPlotW;
  scatter->resize (600, 300);
  sprintf (title,"Folding free energy results");
  scatter->setCaption (title);
  scatter->show(); 
  gQApp->setMainWidget (scatter);
  gQApp->exec();
  return NULL;
}

#endif

// Other functions ------------------------------------------------------------

static void ct_output(FILE *ofp, char *seq, SQINFO sqinfo, int *ss, int j, int d);
void ShuffleMonoSequence (char *seq, int len, int endpos);
void ShuffleDiSequence (char *seq, int len, int endpos);
void ShuffleTriSequence (char *seq, int len, int endpos);

// Main -----------------------------------------------------------------------

int main (int argc, char* argv[])
{
#ifdef HAVE_LIBQT
  vector< DataOutput<char*>* > *gDFFE;
  int serie_id_curr;
  vector< DataOutput<char*>* > *gDFFE_previous;
  int serie_id_prev;
#endif

  int i, j;
  int shuffle_nb = 500;        // Nb of shuffle
  int overlap_size = 150;      // Overlap size < Window size
  int window_size = 200;       // Window size <= 200
  bool ctoutput = false;
  bool shuffle_dinuc = false;
  bool shuffle_trinuc = false;
  int shift_trinuc = 0;

  // Parameters for pknot
  struct rnapar_2 zkn_param;    /* rna parameters secon order + knots              */ 
  int **icfg;                   /* integer log form grammar for alignment          */
  int                   *ss;    /* secondary structure (should go to SQINFO)       */
  float            cykpairs;    /* number of base paired of predicted structure    */
  float               pairs;    /* number of base paired as input                  */

  bool allow_pseudoknots = false;
  bool allow_coaxials = false;
  bool verbose = false;
  bool traceback = false;       /* TRUE print traceback                            */
  struct tracekn_s *tr;	        /* traceback of a predicted RNA structure          */
  int score;                    /* score of predicted structure                    */
  int approx = false;           /* TRUE  == external pseudoknot approximation      *
				 *          exclude diagrams V7-V10 and WB9-WB10   *
				 * FALSE == full pseudoknot model                  *
				            include diagrams V7-V10 and WB9-WB10   */

  int format;            
  SQFILE *sqf;             // file	
  SQINFO sqinfo;           // info structures for seq
  FILE *pknot_file = NULL; // outputfile
  char *seq;	           // seq
  char *orig_seq;
  vector< pair< char*, SQINFO > > windows; 

  // Parse the option line
  actual_name = argv[0];
  if (argc==1) { __usage (); exit (EXIT_FAILURE); }
  while ((__current_option = getopt (argc, argv, __shortopts)) != EOF) 
    {
      switch (__current_option) {
      case 'd':
	shuffle_dinuc = true;
	break;
      case 't':
	shuffle_trinuc = true;
	shift_trinuc = atoi (optarg);
	break;
      case 'g':
	ctoutput = true;
	break;
      case 'k':
	allow_pseudoknots = true;
	break;
      case 'l':
	overlap_size = atoi (optarg);
	break;
      case 's':
	shuffle_nb = atoi (optarg);
	break;
      case 'w':
	window_size = atoi (optarg);
	break;
      case 'h': case 'H':
	__usage (); __help (); 
	exit (EXIT_SUCCESS); 
	break;
      case 'v':
	verbose = true; 
	break;
      case 'V':  
	__version (); __usage (); 
	exit (EXIT_SUCCESS); 
	break;
      default: 
	__usage (); exit (EXIT_FAILURE);
      }
    }
  
  if (argc - optind != 1 ) { __usage (); exit (EXIT_FAILURE); }

  __version ();

#ifdef HAVE_LIBQT
  // Create QApplication
  gQApp = CreateQApplication ();
#endif

  // Do some check!
  if (shuffle_nb <0) {
    cerr << "The number of shuffling must be greater than 0" << endl;
    return EXIT_FAILURE;
  }
  if (window_size <= overlap_size) {
    cerr << "The overlap of window (l) must be less than the window size (w)" << endl;
    return EXIT_FAILURE;
  }
  if (shuffle_dinuc && shuffle_trinuc) {
    cerr << "Switches -d and -t conflicts, please use only one" << endl;
    return EXIT_FAILURE;
  }


  // Let's use the squid library to parse the input file!
  if (! SeqfileFormat(argv[optind], &format, NULL)) {
    cerr << "Error opening file " << argv[optind] 
	 << " : Format not recognized." << endl;
    return EXIT_FAILURE;
  }
  if ((sqf = SeqfileOpen(argv[optind], format, NULL)) == NULL) {
    cerr << "Error opening file " << argv[optind]  << endl;
    return EXIT_FAILURE;
  }
  
  // Job!
  Parameters2_Zkn(&zkn_param);
  icfg = ParamIntSCFG(&zkn_param); 
  
  cout << "Treating sequences from file: " << argv[optind] << endl;
  cout << "DFFE output done in: " << "dffe.out" << endl;
  
  if (verbose) {
    cout << "Window size          = " << window_size << endl;
    cout << "Overlap of window    = " << overlap_size << endl;
    cout << "Number of random ffe = " << shuffle_nb << endl;
  }

  while (ReadSeq(sqf, format, &seq, &sqinfo)) {
    s2upper(seq);
    StripDegeneracy(seq);
    if (sqinfo.flags & SQINFO_SS) free(sqinfo.ss);
    
    orig_seq = seq;

    // Shift the beginning of the sequence
    if (shuffle_trinuc) {
      seq += shift_trinuc;
      sqinfo.len -= shift_trinuc;
      sqinfo.start += shift_trinuc;
      sqinfo.flags = SQINFO_LEN | SQINFO_START | SQINFO_STOP;
    }

    // Extracts windows
    cout << "Extracting all windows of size " << window_size << "..." << flush;
    i=0;
    while (sqinfo.len-i >= window_size) {
      char* tmp = new char[window_size+1];
      tmp[window_size] = '\0';
      SQINFO tmpInfo;
      tmpInfo.flags = SQINFO_LEN | SQINFO_START | SQINFO_STOP;
      tmpInfo.len = window_size;
      tmpInfo.start = sqinfo.start + i;
      tmpInfo.stop = sqinfo.start + i + window_size -1;
      strncpy (tmp, &seq[i], window_size);
      windows.push_back (make_pair (tmp, tmpInfo));
      i+=window_size-overlap_size;
    }

    if (windows.size ()==0) {
      cerr << "Sequence too short." << endl;
      exit (EXIT_FAILURE);
    } else {
      cout << " found " << windows.size () << "." << endl;
    }


#ifdef HAVE_LIBQT
    // Set scatter plot
    gDFFE = new vector< DataOutput< char* >* >;
    gDFFE_previous = new vector< DataOutput< char* >* >;
#endif

    // Find FFE on each windows
    float ffe = 0.0;
    float ffe1;
    vector< float > ffes;
    vector< float > sffes;
    char ffe_filename[80];
    char pknot_filename[80];

    cout << "Calculating FFE:" << endl;
    for (i=0; i<windows.size (); ++i) {
      sprintf (pknot_filename, "window_%05d-%05d.ss", windows[i].second.start, windows[i].second.stop);
      sprintf (ffe_filename, "window_%05d-%05d.out", windows[i].second.start, windows[i].second.stop);
      ofstream ffe_file (ffe_filename);
      
      // Output file
      if ((pknot_file = fopen(pknot_filename, "w")) == NULL) {
	cerr << "Error opening pknot output file " << endl;
	return EXIT_FAILURE;
      }
      
      cout << "window " << windows[i].second.start << "-" 
	   << windows[i].second.stop << ": " << flush;
      StructurePredictkn_2IS(pknot_file, windows[i].first, windows[i].second.len, icfg, 
			     verbose, traceback, &tr, &score, allow_coaxials, 
			     allow_pseudoknots, approx);
      Tracekn(tr, windows[i].first, windows[i].second.len, false, &(windows[i].second.ss));
      Traceintkn(tr, windows[i].first, windows[i].second.len, false, &ss);
      windows[i].second.flags |= SQINFO_SS;
      WriteSeqkn(pknot_file, windows[i].first, &windows[i].second, ss, format, &pairs, &cykpairs);
      if (ctoutput)
	ct_output(pknot_file, windows[i].first, windows[i].second, ss, 
		  windows[i].second.len-1, windows[i].second.len-1);
      
      // Add simple PKnot output...
      ffe = -(float)(score)/(INTSCALE*wsf);
      ffes.push_back (ffe);
      sffes.push_back (0.0);
      cout << ffe << " kcal/mol" << endl;
      ffe_file << ffe << endl;
      if (pknot_file != NULL) fclose(pknot_file);
      ffe_file.close ();
    }
    
    cout << "Evaluate AFFE:" << endl;
    cout << "Shuffling and folding " << shuffle_nb << " times..." << endl;

#ifdef HAVE_LIBQT
    serie_id_curr = scatter->AddSerie (QPen (QColor(200,0,0)));
    serie_id_prev = scatter->AddSerie (QPen (QColor(200,180,180)));
    for (i=0; i<windows.size (); ++i) {
      char* window_name = new char[80];
      sprintf (window_name, "window_%05d-%05d.out", windows[i].second.start, windows[i].second.stop);
      gDFFE_previous->push_back (new DataOutput<char*> (window_name, i, 0));
      scatter->AddData ( (*gDFFE_previous)[i], serie_id_prev );
      gDFFE->push_back (new DataOutput<char*> (window_name, i, 0));
      scatter->AddData ( (*gDFFE)[i], serie_id_curr );
    }
#endif
    
    for (j=0; j<shuffle_nb; ++j) {
      cout << "Iteration " << j << "... " << flush;
      for (i=0; i<windows.size (); ++i) {
	sprintf (ffe_filename, "window_%05d-%05d.out", windows[i].second.start, windows[i].second.stop);
	ofstream ffe_file (ffe_filename, ios::app);

	if (shuffle_trinuc)
	  ShuffleTriSequence(windows[i].first, windows[i].second.len, 
			     windows[i].second.len-1);
        else if (shuffle_dinuc)
	  ShuffleDiSequence(windows[i].first, windows[i].second.len, 
			    windows[i].second.len-1);
	else
	  ShuffleMonoSequence(windows[i].first, windows[i].second.len, 
			    windows[i].second.len-1);

	StructurePredictkn_2IS(pknot_file, windows[i].first, windows[i].second.len, icfg, 
			       verbose, traceback, &tr, &score, allow_coaxials, 
			       allow_pseudoknots, approx);
	ffe1 = -(float)(score)/(INTSCALE*wsf);
	sffes[i] += ffe1;
#ifdef HAVE_LIBQT
	(*(*gDFFE_previous)[i])[1] = (*(*gDFFE)[i])[1];
	(*(*gDFFE)[i])[1] = ffes[i]- sffes[i]/(j+1);
	scatter->SetDirty ();
#endif
	ffe_file << ffe1 << endl;
	ffe_file.close ();
      }

      // Output of results for this iteration in dffe_file
      ofstream dffe_file ("dffe.out");
      for (i=0; i<windows.size (); ++i) {
	sprintf (ffe_filename, "window_%05d-%05d.out", windows[i].second.start, windows[i].second.stop);
	dffe_file << ffe_filename << " " << i << " " << ffes[i] - sffes[i]/(j+1) << endl;
      }
      dffe_file.close ();
      cout << "done." << endl;
    }
    
    // Release some objects
    FreeTracekn(tr);
    FreeSequence(seq, &sqinfo);
    delete[] orig_seq;

    if (windows.size ()>0) {
      for (i=0; i<windows.size (); ++i) {
	FreeSequence(windows[i].first, &windows[i].second);
      }
      windows.clear ();
    }
  }
  
  FreeIntSCFG(icfg);
  SeqfileClose(sqf);

  cout << "All done." << endl;

#ifdef HAVE_LIBQT
  // Wait for the end of other threads
  while (pthread_join (theThread, NULL)) sleep(1);
#endif

  return EXIT_SUCCESS;
}



void ShuffleMonoSequence (char *seq, int len, int endpos)
{
  float *afterfreq;
  float *beforefreq; 
  int i;

  BaseFreq(seq, endpos, len-1, &beforefreq);

  random_shuffle (seq, seq+len);
  
  BaseFreq(seq, endpos, len-1, &afterfreq);

  for (i=0; i<4; ++i)
    if (beforefreq[i] != afterfreq[i]) Die ("wrong shuffling");
}



void ShuffleDiSequence (char *seq, int len, int endpos)
{
  float *beforefreq; 
  float *afterfreq;
  char trinuc[3];
  char basepool[4];
  int i, j, k;

  basepool[0] = 'A'; basepool[1] = 'C'; 
  basepool[2] = 'G'; basepool[3] = 'U'; 

  BaseFreq(seq, endpos, len-1, &beforefreq);

  // Initialize random seed
  srand ((int)time(NULL));

  // Shuffle a lot of time!
  int x;
  for (x=0; x<len*10; ++x) {

    // Select the random basepool
    random_shuffle (basepool, basepool+4);
    trinuc[0] = basepool[0];
    //random_shuffle (basepool, basepool+4);
    //trinuc[1] = basepool[0];
    random_shuffle (basepool, basepool+4);
    trinuc[2] = basepool[0];
    
    // find all occurences of non-overlapping trinucleotides and keep
    // the position of the middle nucleotide
    int i;
    vector< char* > ptrs;
    i = 0;
    while (i < len-2) {
      if (seq[i] == trinuc[0] && seq[i+2] == trinuc[2]) {
	ptrs.push_back (&seq[i+1]);
	i+=3;
      } else {
	i++;
      }
    }
    
    // swap found trinucleotides between them for len times.
    for (j=0; j<len; ++j) {
      for (k=0; k<ptrs.size (); ++k) {
	int l = (int)((float)ptrs.size ()*rand ()/(RAND_MAX+1.0));
	char tmp;
	// swap the two trinuc
	tmp = *ptrs[k];
	*ptrs[k] = *ptrs[l];
	*ptrs[l] = tmp;
      }
    }
    
    ptrs.clear ();
  }

  BaseFreq(seq, endpos, len-1, &afterfreq);

  for (i = 0; i < 4; i++)
    if (beforefreq[i] != afterfreq[i]) Die ("wrong shuffling");
}

void ShuffleTriSequence (char *seq, int len, int endpos)
{
  vector< char* > ptrs;
  int i,j;
  float *afterfreq;
  float *beforefreq; 

  BaseFreq(seq, endpos, len-1, &beforefreq);

  //  cout << "i = " << seq << endl;

  char *newSeq = new char[len+1];
  memcpy (newSeq, seq, len+1);

  i=0;
  while (i < len-2) {
    ptrs.push_back (&newSeq[i]);
    i+=3;
  }
    
  random_shuffle (ptrs.begin (), ptrs.end ());

  for (i=0, j=0; i<ptrs.size (); ++i) {
    seq[j++] = *ptrs[i];
    seq[j++] = *(ptrs[i]+1);
    seq[j++] = *(ptrs[i]+2);
  }
  
  BaseFreq(seq, endpos, len-1, &afterfreq);

  for (i = 0; i < 4; i++)
    if (beforefreq[i] != afterfreq[i]) Die ("wrong shuffling");

  //  cout << "f = " << seq << endl;

  delete[] newSeq;
}





/* Function: ct_output()
 * 
 * Purpose:  Converts the output of pknots program to the "connect" format.
 *           It allows to convert any fragment of the sequence.
 *           
 * Args:     outf     - where to send the output
 *           seq      - sequence to fold. 
 *           sqinfo   - info structures for seq                 
 *           ss       - secondary structure in integer form.
 *                      ss[i] = j  if position i paired to position j
 *                      ss[i] = -1 if position i is unpaired
 *           j        - final position of fragment to report. 
 *                      j\in [0,seqlen-1], where seqlen is the total length of the sequence.
 *           d        - length of the fragment.
 *         
 * Return:   (void)
 */          
static void
ct_output(FILE *ofp, char *seq, SQINFO sqinfo, int *ss, int j, int d)
{  
  int i;          /* initial position of fragment. 0,..,d-1 */
  int iabs;       /* absolute value of initial position     */

  fprintf(ofp,"\n ct_output \n");
  fprintf(ofp,"----------------------------------------------------------------------\n");

  for (i = 0; i <= d; i++) {
    iabs = i + j - d;
    fprintf(ofp, "%5d %c   %5d %4d %4d %4d\n",
            i+1, seq[iabs], i, i+2, (ss[iabs] != -1)? ss[iabs]+1-j+d:0, iabs+1);
  }
  
  fprintf(ofp, "\n");
  
}      
