/* rnaoutput.c
 * derived from COVE's konings.c
 * SRE, Sun Aug 28 10:39:18 1994
 * 
 * Representation of secondary structure and secondary structural 
 * alignments using a variant of Danielle Konings' string notation.
 * 
 * See: Konings and Hogeweg, J. Mol. Biol. 207:597-614 1989
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "cfg.h"
#include "proto.h"
#include "protovx.h"
#include "protowbx.h"
#include "protowx.h"
#include "squid.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif


/* Function: Tracekn()
 * 
 * Purpose:  From a traceback tree of seq, produce a
 *           secondary structure string. ">" and "<" are
 *           used for pairwise emissions; "." for single-stranded stuff.
 *           Note that structure is defined by pairwise emissions,
 *           not by Watson-Crick-isms and stacking rules.
 *
 *           Constructed for the one-hole algorithm.
 *           
 * Args:     tr          - traceback structure
 *           seq         - sequence, 0..rlen-1
 *           rlen        - length of seq and returned ss string
 *           watsoncrick - TRUE to annotate canonical pairs only
 *           ret_ss      - RETURN: alloc'ed secondary structure string
 *
 * Return:   ret_ss contains a string 0..rlen-1 containing the
 *           secondary structure. Must be free'd by caller.
 */
void
Tracekn(struct tracekn_s *tr, char *seq, int rlen, int watsoncrick, 
        char **ret_ss)  
{ 
  struct traceknstack_s *dolist;
  struct tracekn_s      *curr;
  char                  *ss;

  if ((ss = (char *) malloc (sizeof(char) * rlen+1)) == NULL)
    Die("malloc failed");
  memset(ss, '.', rlen);
  ss[rlen] = '\0';

  dolist = InitTraceknstack();
  PushTraceknstack(dolist, tr->nxtl);

  while ((curr = PopTraceknstack(dolist)) != NULL)
    {
      if (curr->type1 == dpcP || curr->type1 == dpcPS || curr->type1 == dpcPI || curr->type1 == dpcPL)
	{
	  if (! watsoncrick  ||
	      IsRNAComplement(seq[curr->emiti], seq[curr->emitj], TRUE))
	    {
	      ss[curr->emiti] = '>';
	      ss[curr->emitj] = '<';
	    }
	}

      if (curr->type2 == dpcP || curr->type2 == dpcPS || curr->type2 == dpcPI || curr->type2 == dpcPL)
	{
	  if (! watsoncrick  ||
	      IsRNAComplement(seq[curr->emitk], seq[curr->emitl], TRUE))
	    {
	      ss[curr->emitk] = '>';
	      ss[curr->emitl] = '<';
	    }
	}

      if (curr->nxtr) PushTraceknstack(dolist, curr->nxtr);
      if (curr->nxtl) PushTraceknstack(dolist, curr->nxtl);
    }
  FreeTraceknstack(dolist);
  *ret_ss = ss;
}

/* Function: Traceintkn()
 * 
 * Purpose:  Convert a secondary structure string to an array of integers
 *           representing what position each position is base-paired 
 *           to (0..len-1), or -1 if none. 
 *           
 *           Constructed for the one-hole algorithm.
 *
 *           "." are used for single-stranded stuff.
 *           Note that structure is defined by pairwise emissions,
 *           not by Watson-Crick-isms and stacking rules.
 *           
 * Args:     tr          - traceback structure
 *           seq         - sequence, 0..rlen-1
 *           rlen        - length of seq and returned ss string
 *           watsoncrick - TRUE to annotate canonical pairs only
 *           ret_ss      - RETURN: alloc'ed secondary structure string
 *
 * Return:   ret_ss contains a string 0..rlen-1 containing the
 *           secondary structure. Must be free'd by caller.
 */
void
Traceintkn(struct tracekn_s *tr, char *seq, int rlen, int watsoncrick, 
	   int **ret_ss)  
{ 
  struct traceknstack_s *dolist;
  struct tracekn_s      *curr;
  int                   *ss;
  int                    i;

  if ((ss = (int *) malloc (sizeof(int) * rlen+1)) == NULL)
    Die("malloc failed");

  for (i = 0; i < rlen; i++)
    ss[i] = -1;

  dolist = InitTraceknstack();
  PushTraceknstack(dolist, tr->nxtl);

  while ((curr = PopTraceknstack(dolist)) != NULL)
    {
      if (curr->type1 == dpcP || curr->type1 == dpcPS || curr->type1 == dpcPI || curr->type1 == dpcPL)
	{
	  if (! watsoncrick  ||
	      IsRNAComplement(seq[curr->emiti], seq[curr->emitj], TRUE))
	    {
	      ss[curr->emiti] = curr->emitj;
	      ss[curr->emitj] = curr->emiti;
	    }
	}


      if (curr->type2 == dpcP || curr->type2 == dpcPS || curr->type2 == dpcPI || curr->type2 == dpcPL )
	{
	  if (! watsoncrick  ||
	      IsRNAComplement(seq[curr->emitk], seq[curr->emitl], TRUE))
	    {
	      ss[curr->emitk] = curr->emitl;
	      ss[curr->emitl] = curr->emitk;
	    }
	}

      if (curr->nxtr) PushTraceknstack(dolist, curr->nxtr);
      if (curr->nxtl) PushTraceknstack(dolist, curr->nxtl);
    }
  FreeTraceknstack(dolist);
  *ret_ss = ss;
}

/* Function: WriteSeqkn()
 * 
 * Purpose:  writes in outf file the Tracekn
 *           
 *           Constructed for the one-hole algorithm.
 *
 *           "." are used for single-stranded stuff.
 *           Note that structure is defined by pairwise emissions,
 *           not by Watson-Crick-isms and stacking rules.
 *           
 */

int
WriteSeqkn(FILE *outf, char *seq, SQINFO *sqinfo, int *ss, int format, float *ret_pairs, float *ret_cykpairs)
{
  int   numline = 0;
  int   lines = 0, spacer = 4, width = 20, tab = 0;
  int   i, j, l, l1, ibase, m;
  char  endstr[10]; 
  char  s[100];			/* buffer for sequence  */
  int   pos[100];		/* buffer for structure */
  int   lss[100];		/* buffer for secondary structure */
  int   checksum = 0;
  int   seqlen;   
  int   dostruc;		/* TRUE to print structure lines*/
  int   pairs = 0;
  int   cykpairs = 0;

  dostruc    = FALSE;		
  seqlen     = (sqinfo->flags & SQINFO_LEN) ? sqinfo->len : strlen(seq);

  strcpy( endstr,"");
  l1 = 0;

  /* 10Nov91: write this out in all possible formats: */
  checksum = GCGchecksum(seq, seqlen);

  fprintf(outf, "NAM  %s\n", sqinfo->name);

  if (sqinfo->flags & (SQINFO_ID | SQINFO_ACC | SQINFO_START | SQINFO_STOP | SQINFO_OLEN))
    fprintf(outf, "SRC  %s %s %d..%d::%d\n",
	    (sqinfo->flags & SQINFO_ID)    ? sqinfo->id     : "-",
	    (sqinfo->flags & SQINFO_ACC)   ? sqinfo->acc    : "-",
	    (sqinfo->flags & SQINFO_START) ? sqinfo->start  : 0,
	    (sqinfo->flags & SQINFO_STOP)  ? sqinfo->stop   : 0,
	    (sqinfo->flags & SQINFO_OLEN)  ? sqinfo->olen   : 0);

  if (sqinfo->flags & SQINFO_DESC)
    fprintf(outf, "DES  %s\n", sqinfo->desc);

  if (sqinfo->flags & SQINFO_SS) {
    fprintf(outf, "SEQ  +SS\n");
    dostruc = TRUE;	/* print structure lines too */
  }
  else
    fprintf(outf, "SEQ\n");

  numline = 1;                /* number seq lines w/ coords  */
  strcpy(endstr, "\n");

  if(format == kSquid || format == kSelex)
    for (i = 0; i < seqlen; i++) 
      if (sqinfo->ss[i] != '.') 
	pairs += 1;
  
  for (i=0, l=0, ibase = 1, lines = 0; i < seqlen; ) {

    if (l1 < 0) 
      l1 = 0;
    else if (l1 == 0) {
      if (numline) 
	fprintf(outf,"%8d ",ibase);
      for (j=0; j<tab; j++) 
	fputc(' ',outf);
    }
      
    if (spacer != 0 && l%spacer == 1) {
      s[l] = ' '; 
      lss[l] = 1234; 
      l++;
    }
    
    if (spacer != 0 && l%spacer == 2) {
      s[l] = ' '; 
      lss[l] = 1234; 
      l++;
    }
      
    if (spacer != 0 && l%spacer == 3) {
      s[l] = ' '; 
      lss[l] = 1234;
      l++;
    }

    pos[l] = i;
    s[l]   = *(seq+i);
      
    if (ss[i] != -1) {
      lss[l]  = ss[i];
      cykpairs += 1;
    }
    else 
      lss[l] = 56789;
    
    l++; i++;
    l1++;                 /* don't count spaces for width*/
    if (l1 == width || i == seqlen) {
      s[l]  = '\0';
      lss[l] = 888888;
      
      if (dostruc) {
	fprintf(outf, "%s\n", s);
	
	if (numline) 
	  fprintf(outf,"         ");
	
	for (j=0; j<tab; j++) 
	  fputc(' ',outf);
	
	for (m=0; m<l; m++)
	  if (s[m] != ' '  &&  pos[m] <= 9) 
	    fprintf(outf,"%d   ", *(pos+m));
	  else if (s[m] != ' '  && pos[m] > 9 && pos[m] <= 99) 
	    fprintf(outf,"%d  ", *(pos+m));
	  else if (s[m] != ' ') 
	    fprintf(outf,"%d ", *(pos+m));
	
	fprintf(outf,"\n");
	
	if (numline) 
	  fprintf(outf,"         ");
	
	for (j = 0; j < tab; j++) 
	  fputc(' ',outf);
	
	for (m = 0; m < l; m++)
	  if (s[m] != ' '  && lss[m] <= 9 && lss[m] != 56789)
	    fprintf(outf,"%d   ", *(lss+m));
	  else if (s[m] != ' '  && lss[m] > 9 && lss[m] <= 99 && lss[m] != 56789)
	    fprintf(outf,"%d  ", *(lss+m));
	  else if (s[m] != ' ' && lss[m] != 56789)
	    fprintf(outf,"%d ", *(lss+m));
	  else if (s[m] != ' ') 
	    fprintf(outf, ".   ");
	
	fprintf(outf,"%s\n",endstr);
      }
      else {
	if (i == seqlen) fprintf(outf,"%s%s\n",s,endstr);
	else fprintf(outf,"%s\n",s);
      }
      l = 0; l1 = 0;
      lines++;
      ibase = i+1;
    }
  }

  *ret_pairs    = pairs/2;
  *ret_cykpairs = cykpairs/2;

  return lines;
} 

/* Function: IsRNAComplement()
 * 
 * Purpose:  Returns TRUE if sym1, sym2 are Watson-Crick complementary.
 *           If allow_gu is TRUE, GU pairs also return TRUE.
 */
int
IsRNAComplement(char sym1, char sym2, int allow_gu)
{
  sym1 = toupper(sym1);
  sym2 = toupper(sym2);

  if ((sym1 == 'A' && sym2 == 'U') ||
      (sym1 == 'C' && sym2 == 'G') ||
      (sym1 == 'G' && sym2 == 'C') ||
      (sym1 == 'U' && sym2 == 'A') ||
      (allow_gu && sym1 == 'G' && sym2 == 'U') ||
      (allow_gu && sym1 == 'U' && sym2 == 'G'))
    return TRUE;
  else
    return FALSE;
}

void
CalculatePairs(SQINFO sqinfo, int *ss, int format, float *ret_pairs, float *ret_cykpairs)
{
  int i;
  float pairs = 0.;
  float cykpairs = 0.;

  for (i = 0; i < sqinfo.len; i++) {
    if ((format == kSquid || format == kSelex ) && sqinfo.ss[i] != '.') 
      pairs += 1.;
    if (ss[i] != -1) 
      cykpairs += 1.;
  }

  *ret_pairs    = pairs/2.;
  *ret_cykpairs = cykpairs/2.;
}


/* Function: VerifyKHS()
 * 
 * Purpose:  Examine a possibly bad structure string, and print out diagnostics 
 *           about it if wordy is TRUE.
 *
 * Return:   1 if string is OK, 0 if string is bad.
 */
int
VerifyKHS(char *name, char *ss, int wordy)
{
  int symcount[27];             /* 0 is normal pairs. 1-26 for pseudoknots */
  int i;
  int pos;
  int status = 1;

  for (i = 0; i < 27; i++)
    symcount[i] = 0;

  for (pos = 0; ss[pos] != '\0'; pos++)
    {
      if (!sre_isascii(ss[pos]))        /* SGI ctype.h non-ANSI compliant */
        {
          status = 0;
          if (wordy)
            Warn("VerifyKHS: Sequence %s no good. structure has garbagesymbol (val %d) at position %d", 
		 name, (int)ss[pos], pos);
        }
      else if (ss[pos] == '>')
        symcount[0] ++;
      else if (ss[pos] == '<')
        symcount[0] --;
      else if (isupper((int) ss[pos]))
        symcount[ss[pos] - 'A' + 1] ++; /* pseudoknot-left  */
      else if (islower((int) ss[pos])) 
        symcount[ss[pos] - 'a' + 1] --; /* pseudoknot-right */
      else if (!isgap(ss[pos]))
        {
          status = 0;
          if (wordy)
            Warn("VerifyKHS: Sequence %s no good, structure has invalid symbol %c at position %d", 
		 name, (int)ss[pos], pos);
        }
    }
      
  if (symcount[0] != 0)
    {
      status = 0;
      if (wordy)
        Warn("VerifyKHS: Sequence %s no good, structure has extra paired bases: %d on %s",
             name, abs(symcount[0]), 
             (symcount[0] > 0) ? "left" : "right");
    }

  for (i = 1; i <= 26; i++)
    if (symcount[i] != 0)
      {
        status = 0;
        if (wordy)
          Warn("VerifyKHS: Sequence %s no good, structure has extra paired bases for pseudoknot %c: %d on %s",
               name, (char) (i + 'A' - 1),
               abs(symcount[i]), 
               (symcount[i] > 0) ? "left" : "right");
      }
  return status;
}





























